''' basic functions and functions for manipulating ENVI binary files'''
import os
import sys
import copy
import math
import struct
import numpy as np
import os.path as path
import matplotlib.pyplot as plt


def err(msg):
    print('Error: ' + msg)
    sys.exit(1)


def run(cmd):
    a = os.system(cmd)
    if a != 0:
        err("command failed: " + cmd)


def exist(f):
    return os.path.exists(f)


def hdr_fn(bin_fn):
    # return filename for header file, given name for bin file
    hfn = bin_fn[:-4] + '.hdr'
    if not exist(hfn):
        hfn2 = bin_fn + '.hdr'
        if not exist(hfn2):
            err("didn't find header file at: " + hfn + " or: " + hfn2)
        return hfn2
    return hfn


def read_hdr(hdr):
    samples, lines, bands = 0, 0, 0
    for line in open(hdr).readlines():
        line = line.strip()
        words = line.split('=')
        if len(words) == 2:
            f, g = words[0].strip(), words[1].strip()
            if f == 'samples':
                samples = g
            if f == 'lines':
                lines = g
            if f == 'bands':
                bands = g
    return samples, lines, bands


# use numpy to read floating-point data, 4 byte / float, byte-order 0
def read_float(fn):
    return np.fromfile(fn, '<f4')


def wopen(fn):
    f = open(fn, "wb")
    if not f:
        err("failed to open file for writing: " + fn)
    return f


def read_binary(fn):
    hdr = hdr_fn(fn)

    # read header and print parameters
    samples, lines, bands = read_hdr(hdr)

    data = read_float(fn)
    return samples, lines, bands, data


def write_binary(np_ndarray, fn):
    of = wopen(fn)
    np_ndarray.tofile(of, '', '<f4')
    of.close()


def write_hdr(hfn, samples, lines, bands):
    lines = ['ENVI',
             'samples = ' + str(samples),
             'lines = ' + str(lines),
             'bands = ' + str(bands),
             'header offset = 0',
             'file type = ENVI Standard',
             'data type = 4',
             'interleave = bsq',
             'byte order = 0']
    open(hfn, 'wb').write('\n'.join(lines).encode())


def hist(data):
    # counts of each data instance
    count = {}
    for d in data:
        count[d] = 1 if d not in count else count[d] + 1
    return count


def parfor(my_function, my_inputs):
    # evaluate a function in parallel, and collect the results
    import multiprocessing as mp
    pool = mp.Pool(mp.cpu_count())
    result = pool.map(my_function, my_inputs)
    return(result)



import numpy as np
import  matplotlib.pyplot as plt
from rasterbox.utilities import Misc
from rasterbox.utilities import DataManip


class Image(object):
    def __init__(self, bin=None, nparray=None):
        if nparray is None:
            samples, lines, bands, data = Misc.read_binary(bin)
            self.samples, self.lines, self.bands = \
                int(samples), int(lines), int(bands)
            self._Data = data.reshape(self.lines * self.samples, self.bands)
            self.__build_rgb()
        else:
            self._Data = nparray


    def __build_rgb(self):
        arr = np.zeros((self.lines, self.samples, 3))

        # data has to switch around for matplotlib
        tmp = self._Data.reshape(self.bands, self.lines * self.samples)
        for i in range(0, 3):
            arr[:, :, i] = tmp[3 - i, :].reshape((self.lines, self.samples))

        for i in range(0, 3):
            arr[:, :, i] = DataManip.rescale(arr[:, :, i])

        del tmp
        self.rgb = arr

    def spatial(self):
        return self._Data.reshape((self.samples,self.lines,self.bands))


    def Data(self):
        return self._Data


    def showplot(self):
        plt.imshow(self.rgb)
        plt.tight_layout()
        plt.show()


from rasterbox.utilities import Misc
import matplotlib.pyplot as plt
class Target(object):


    def __init__(self, name, bin, target_list):
        self.name = name
        s,l,b,self.Data = Misc.read_binary(bin)
        self.samples, self.lines, self.bands = int(s), int(l), int(b)
        self.__build_binary(target_list)


    def __build_binary(self, target_list):
        targets = target_list
        if self.name in targets:
            b = targets[self.name]['bool']
            val =  targets[self.name]['val']

            if b:
                self.Binary = self.Data == float(val)
            elif not b:
                self.Binary = self.Data != float(val)
            else:
                raise Exception('There was an error encoding binaries')


    def ravel(self, binary=True):
        if binary:
            return self.Binary.reshape(self.lines * self.samples)
        else:
            return self.Data.reshape(self.lines *  self.samples)


    def spatial(self, binary=True):
        if binary:
            return self.Binary.reshape(self.lines, self.samples)
        else:
            return self.Data.reshape(self.lines, self.samples)


    def showplot(self, binary=True):
        y = self.spatial(binary=binary)
        if binary:
            plt.imshow(y, cmap='gray')
        else:
            plt.imshow(y)
        plt.tight_layout()
        plt.show()


import copy
import math
import numpy as np
from rasterbox.utilities import Misc

"""
######    General use functions
"""
def ravel(lines, samples, bands, arr):
    return arr.reshape(int(lines) * int(samples), int(bands))

def spatial(lines, samples, bands, arr):
    return arr.reshape(int(lines), int(samples), int(bands))

def create_batch_generator(X, y, batch_size=128, shuffle=False):
    X_copy = np.array(X)
    y_copy = np.array(y)

    if shuffle:
        data = np.column_stack((X_copy, y_copy))
        np.random.shuffle(data)
        X_copy = data[:, :-1]
        y_copy = data[:, :-1].astype(int)

    for i in range(0, X.shape[0], batch_size):
        yield (X_copy[i: i+batch_size, :], y_copy[i: i+batch_size])

def rescale(arr, two_percent=True):
    arr_min = arr.min()
    arr_max = arr.max()
    scaled = (arr - arr_min) / (arr_max - arr_min)

    if two_percent:
        # 2%-linear stretch transformation for hi-contrast vis
        values = copy.deepcopy(scaled)
        values = values.reshape(np.prod(values.shape))
        values = values.tolist()
        values.sort()
        npx = len(values)  # number of pixels
        if values[-1] < values[0]:
            print('error: failed to sort')
            sys.exit(1)
        v_min = values[int(math.floor(float(npx)*0.02))]
        v_max = values[int(math.floor(float(npx)*0.98))]
        scaled -= v_min
        rng = v_max - v_min
        if rng > 0.:
            scaled /= (v_max - v_min)

    return scaled


import os
import math
import logging
import numpy as np
from rasterbox.utilities import Misc
from rasterbox.Image import Image
from rasterbox.Target import Target


class Rasterbox(object):
    logging.basicConfig(filename="rasterbox.log", level=logging.INFO)

    def __init__(self, src, images_path, targets_path, target_dict):
        logging.info("Init Rasterbox")
        self.src = src
        self.targets = target_dict
        self.__build(images_path, targets_path)

    def __build(self, images_path, targets_path):
        """
            The build function takes two params - images_path
            and targets_path - each of which should represent the
            parent of folder that these images reside in.

            The local path is then calculated and passed to
            build_binaries, where the local path for the images
            and their subsequent headers are derived.

            Finally, both images and targets are built and stored
            within the object as members in several forms including
            rgb and raveled data.
        """
        images_path = os.path.join(self.src, '%s' % images_path)
        targets_path = os.path.join(self.src, '%s' % targets_path)

        logging.info("Images Path " + images_path)
        logging.info("Targets Path " + targets_path)

        if images_path == None or targets_path == None:
            logging.error("Image or target path could not be located")
            raise FileNotFoundError("Couldn't locate images path or targets path")

        image_bins = self.__build_binaries(images_path)
        target_bins = self.__build_binaries(targets_path)

        if image_bins == None:
            logging.error("There is a problem with the path to binary images")
            raise Exception(images_path + " returns None object")
        if target_bins == None:
            logging.error("There is a problem with the path to binary targets")
            raise Exception(targets_path + " returns None object")

        self.__build_images(image_bins)
        self.__build_targets(target_bins)

        if self.S2 and self.L8:
            self.__build_combined_satellites()
        self.__build_onehot()


    def __build_images(self, bins):
        for idx, bin in enumerate(bins):
            if 'S2' in bin:
                self.S2 = Image(bins[idx])
                logging.info("Sentinel 2 built")

            elif 'L8' in bin:
                self.L8 = Image(bins[idx])
                logging.info("Landsat 8 built")

            else:
                Misc.err("Do not recognize file ", bin)

    def __build_targets(self, bins):
        self.Target = dict()

        for _, bin in enumerate(bins):
            self.Target.update({
                target: Target(target, bin, self.targets)
                for target in self.targets if target in bin.lower()
            })
        logging.info("Targets built")


    def __build_binaries(self,path):
        logging.info("Building binaries " + path)
        try:
            for root, dirs, files in os.walk(path, topdown=False):
                bin_files = [
                    os.path.join(path, '%s' % file)
                    for file in files if '.hdr' not in file
                ]
                if bin_files == None:
                    raise FileNotFoundError()
                return bin_files
        except:
            Misc.err("Error building headers and binaries for %s" % path)


    def __build_combined_satellites(self):
        self.Combined = Image(nparray=np.append(self.S2._Data, self.L8._Data, axis=1))


    def __build_onehot(self):
        tmp = np.zeros((self.S2.lines * self.S2.samples, len(self.Target.keys())))
        for idx, target in enumerate(sorted(self.Target.keys())):
            tmp[:,idx] = self.Target[target].Binary
        self.Onehot = Image(nparray=tmp)


    def onehot_inconsistencies():
        pass



import os
import struct
import numpy as np

class DataTest(object):

    @staticmethod
    def load_mnist(path, kind='train'):
        labels_path = os.path.join(path, '%s-labels-idx1-ubyte' % kind)
        images_path = os.path.join(path, '%s-images-idx3-ubyte' % kind)


        with open(labels_path, 'rb') as lbpath:
            magic, n = struct.unpack('>II', lbpath.read(8))
            labels = np.fromfile(lbpath, dtype=np.uint8)

        with open(images_path, 'rb') as imgpath:
            magic, num, rows, cols = struct.unpack(">IIII", imgpath.read(16))
            images = np.fromfile(
                imgpath, dtype=np.uint8).reshape(len(labels), 784)
            images = ((images / 255) - .5) * 2

        return images, labels

    @staticmethod
    def mean_center_normalize(data):
        mean_vals = np.mean(data, axis=0)
        std_val = np.std(data)
        return mean_vals, std_val




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rasterbox",
    version="0.0.5",
    author="Brad Crump",
    author_email="brad@breadcrumbbuilds.com",
    description="Semantic access of raster images and target labels associated with rasters",
    keywords = ["RASTER","IMAGES", "MACHINE","LEARNING"],
    install_requires=[
        'numpy'
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/breadcrumbbuilds/rasterbox",
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
